"""
Provides the main class for the dbt client.
"""

import base64
import subprocess
import os, signal

from re import L
from time import sleep, time
from typing import Callable, Dict, List, Union
from uuid import uuid1


import requests

DEFAULT_SYNC_SLEEP = 0.5


class RPCError(Exception):
    def __init__(self, response, *args, **kwargs):
        self.response: Dict = response
        super().__init__(*args, **kwargs)


class DbtClient:
    """
    DbtClient constructor.

    Args:
        - host (str): The host of the dbt server.
        - port (int): The port of the dbt server.
        - jsonrpc_version (str, optional): The jsonrpc version of the dbt server.
    """

    def __init__(
        self,
        *,
        host: str = "localhost",
        port: int = 8580,
        jsonrpc_version: str = "2.0",
    ):
        """
        Initializes the dbt client.
        """
        self._host = host
        self._port = port
        self._jsonrpc_version = jsonrpc_version
        self.pid = None
        self.session = requests.Session()
        self._url = f"http://{self._host}:{self._port}/jsonrpc"

    def _request(self, method: str, *, params: dict = None) -> dict:
        """
        Makes a request to the dbt server.
        """
        if params is None:
            params = {}
        data = {
            "jsonrpc": self._jsonrpc_version,
            "method": method,
            "params": params,
            "id": str(uuid1()),
        }
        pid = self._get_pid()

        if method == "run_sql":
            print(f"Method: {method}, send HUP to {self.pid}")
            os.kill(self.pid, signal.SIGHUP)
            sleep(DEFAULT_SYNC_SLEEP)

        response = self.session.post(self._url, json=data)
        return response.json()

    def _get_pid(self) -> int:

        if self.pid:
            return self.pid

        proc1 = subprocess.Popen(["ps", "aux"], stdout=subprocess.PIPE)
        proc2 = subprocess.Popen(
            ["grep", "dbt-rpc"],
            stdin=proc1.stdout,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        proc3 = subprocess.Popen(
            ["grep", "-v", "grep"],
            stdin=proc2.stdout,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        proc4 = subprocess.Popen(
            ["awk", "{print $2}"],
            stdin=proc3.stdout,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        proc1.stdout.close()
        proc2.stdout.close()
        proc3.stdout.close()
        out, err = proc4.communicate()
        self.pid = int(out.strip().decode())

        return self.pid

    def status(self) -> dict:
        """
        Gets the status of the dbt server.

        Docs: https://docs.getdbt.com/reference/commands/rpc#status
        """
        return self._request("status")

    def poll(
        self,
        *,
        request_token: str,
        logs: bool = False,
        logs_start: int = 0,
    ) -> dict:
        """
        Polls the dbt server for the status of a request.

        Docs: https://docs.getdbt.com/reference/commands/rpc#poll
        """
        return self._request(
            "poll",
            params={
                "request_token": request_token,
                "logs": logs,
                "logs_start": logs_start,
            },
        )

    def ps(  # pylint: disable=invalid-name
        self,
        *,
        completed: bool = False,
    ):
        """
        Gets the status of the dbt server.

        Docs: https://docs.getdbt.com/reference/commands/rpc#ps
        """
        return self._request(
            "ps",
            params={
                "completed": completed,
            },
        )

    def kill(
        self,
        *,
        task_id: str,
    ):
        """
        Kills a running request.

        Docs: https://docs.getdbt.com/reference/commands/rpc#kill
        """
        return self._request(
            "kill",
            params={
                "task_id": task_id,
            },
        )

    def _run_sync(
        self,
        method: str,
        *,
        params: dict = None,
        logs: bool = False,
        logs_start: int = 0,
        timeout: int = None,
        timeout_action: Union[str, Callable] = "raise",
        task_tags: dict = None,
    ) -> dict:
        """
        Wrapper for running commands synchronously.
        """

        if params is None:
            params = {}

        if task_tags is None:
            task_tags = {}
        params["task_tags"] = task_tags

        if timeout is not None:
            max_time = time() + timeout
            params["timeout"] = timeout

        if (timeout_action not in ["raise", "return"]) and (
            not callable(timeout_action)
        ):
            raise ValueError("timeout_action must be 'raise', 'return' or a callable")

        response_data = self._request(method, params=params)
        if "result" in response_data:
            if "request_token" in response_data["result"]:
                request_token = response_data["result"]["request_token"]
            else:
                raise RPCError(
                    response_data,
                    f"Unknown response format for sync execution: {response_data}",
                )
        else:
            raise RPCError(
                response_data,
                f"Unknown response format for sync execution: {response_data}",
            )

        state = None
        while True:
            if (timeout is not None) and (time() >= max_time):
                if callable(timeout_action):
                    timeout_action()
                elif timeout_action == "raise":
                    raise Exception("Timed out waiting for response")
                elif timeout_action == "return":
                    return response_data
            response_data = self.poll(
                request_token=request_token,
                logs=logs,
                logs_start=logs_start,
            )
            if "result" in response_data:
                if "state" in response_data["result"]:
                    state = response_data["result"]["state"]
                else:
                    raise RPCError(
                        response_data,
                        f"Unknown response format for sync execution: {response_data}",
                    )
            else:
                raise RPCError(
                    response_data,
                    f"Unknown response format for sync execution: {response_data}",
                )
            if state == "success":
                return response_data
            sleep(DEFAULT_SYNC_SLEEP)

    def cli(
        self,
        command: str,
        *,
        logs: bool = False,
        logs_start: int = 0,
        sync: bool = False,
        task_tags: dict = None,
        timeout: int = None,
        timeout_action: Union[str, Callable] = "raise",
    ):
        """
        Runs a dbt command using CLI syntax

        Docs: https://docs.getdbt.com/reference/commands/rpc#running-a-task-with-cli-syntax

        Args:
            - command (str): The command to run. No need to add `dbt` at the beginning.
            - logs (bool, optional): Whether to return logs. Only applies to sync mode.
            - logs_start (int, optional): The number of logs to skip. Only applies to sync mode.
            - sync (bool, optional): Whether to run the command synchronously.
            - task_tags (dict, optional): Arbitrary key/value pairs to attach to this task. These
                tags will be returned in the output of the poll and ps methods
            - timeout (int, optional): The timeout in seconds.
            - timeout_action (str, optional): The action to take if the command times out. Only
                applies to sync mode.
        """
        command = command if not command.startswith("dbt") else command[3:].strip()
        params = {"cli": command}
        if timeout:
            params["timeout"] = timeout
        if task_tags:
            params["task_tags"] = task_tags

        if sync:
            return self._run_sync(
                "cli_args",
                params=params,
                logs=logs,
                logs_start=logs_start,
                timeout=timeout,
                timeout_action=timeout_action,
                task_tags=task_tags,
            )
        return self._request("cli_args", params=params)

    def compile(
        self,
        *,
        exclude: List[str] = None,
        logs: bool = False,
        logs_start: int = 0,
        select: List[str] = None,
        selector: str = None,
        state: str = None,
        sync: bool = False,
        task_tags: dict = None,
        threads: int = None,
        timeout: int = None,
        timeout_action: Union[str, Callable] = "raise",
    ):
        """
        Compiles a project.

        Docs: https://docs.getdbt.com/reference/commands/rpc#compile-a-project-docs

        Args:
            - exclude (list, optional): A list of resources to exclude from compiling, running,
                testing, seeding, or snapshotting .
            - logs (bool, optional): Whether to return logs. Only applies to sync mode.
            - logs_start (int, optional): The number of logs to skip. Only applies to sync mode.
            - select (list, optional): A list of resources to execute.
            - selector (str, optional): The name of a predefined YAML selector that defines the set
                of resources to execute.
            - state (str, optional): The filepath of artifacts to use when establishing
                [state](https://docs.getdbt.com/docs/guides/understanding-state).
            - sync (bool, optional): Whether to run the command synchronously.
            - task_tags (dict, optional): Arbitrary key/value pairs to attach to this task. These
                tags will be returned in the output of the poll and ps methods.
            - threads (int, optional): The number of threads to use.
            - timeout (int, optional): The timeout in seconds.
            - timeout_action (str, optional): The action to take if the command times out. Only
                applies to sync mode.
        """

        exclude_str = " ".join(exclude) if exclude else None
        select_str = " ".join(select) if select else None
        task_tags = task_tags or {}

        params = {}
        if exclude_str:
            params["exclude"] = exclude_str
        if select_str:
            params["select"] = select_str
        if selector:
            params["selector"] = selector
        if state:
            params["state"] = state
        if task_tags:
            params["task_tags"] = task_tags
        if threads:
            params["threads"] = threads
        if timeout:
            params["timeout"] = timeout

        if sync:
            return self._run_sync(
                "compile",
                params=params,
                logs=logs,
                logs_start=logs_start,
                timeout=timeout,
                timeout_action=timeout_action,
                task_tags=task_tags,
            )
        return self._request("compile", params=params)

    def run(
        self,
        *,
        defer: bool = None,
        exclude: List[str] = None,
        logs: bool = False,
        logs_start: int = 0,
        select: List[str] = None,
        selector: str = None,
        state: str = None,
        sync: bool = False,
        task_tags: dict = None,
        threads: int = None,
        timeout: int = None,
        timeout_action: Union[str, Callable] = "raise",
    ):
        """
        Runs models.

        Docs: https://docs.getdbt.com/reference/commands/rpc#run-models-docs

        Args:
            - defer (bool, optional): Whether to defer references to upstream, unselected resources.
                Requires `state`.
            - exclude (list, optional): A list of resources to exclude from compiling, running,
                testing, seeding, or snapshotting .
            - logs (bool, optional): Whether to return logs. Only applies to sync mode.
            - logs_start (int, optional): The number of logs to skip. Only applies to sync mode.
            - select (list, optional): A list of resources to execute.
            - selector (str, optional): The name of a predefined YAML selector that defines the set
                of resources to execute.
            - state (str, optional): The filepath of artifacts to use when establishing
                [state](https://docs.getdbt.com/docs/guides/understanding-state).
            - sync (bool, optional): Whether to run the command synchronously.
            - task_tags (dict, optional): Arbitrary key/value pairs to attach to this task. These
                tags will be returned in the output of the poll and ps methods.
            - threads (int, optional): The number of threads to use.
            - timeout (int, optional): The timeout in seconds.
            - timeout_action (str, optional): The action to take if the command times out. Only
                applies to sync mode.
        """

        if defer is not None and state is None:
            raise Exception("`defer` requires `state` to be set")

        exclude_str = " ".join(exclude) if exclude else None
        select_str = " ".join(select) if select else None
        task_tags = task_tags or {}

        params = {}
        if defer is not None:
            params["defer"] = defer
        if exclude_str:
            params["exclude"] = exclude_str
        if select_str:
            params["select"] = select_str
        if selector:
            params["selector"] = selector
        if state:
            params["state"] = state
        if task_tags:
            params["task_tags"] = task_tags
        if threads:
            params["threads"] = threads
        if timeout:
            params["timeout"] = timeout

        if sync:
            return self._run_sync(
                "run",
                params=params,
                logs=logs,
                logs_start=logs_start,
                timeout=timeout,
                timeout_action=timeout_action,
                task_tags=task_tags,
            )
        return self._request("run", params=params)

    # pylint: disable=too-many-arguments
    def test(
        self,
        *,
        data: bool = True,
        exclude: List[str] = None,
        logs: bool = False,
        logs_start: int = 0,
        schema: bool = True,
        select: List[str] = None,
        selector: str = None,
        state: str = None,
        sync: bool = False,
        task_tags: dict = None,
        threads: int = None,
        timeout: int = None,
        timeout_action: Union[str, Callable] = "raise",
    ):
        """
        Run tests.

        Docs: https://docs.getdbt.com/reference/commands/rpc#run-tests-docs

        Args:
            - data (bool, optional): If True, run data tests.
            - exclude (list, optional): A list of resources to exclude from compiling, running,
                testing, seeding, or snapshotting .
            - logs (bool, optional): Whether to return logs. Only applies to sync mode.
            - logs_start (int, optional): The number of logs to skip. Only applies to sync mode.
            - schema (bool, optional): If True, run schema tests.
            - select (list, optional): A list of resources to execute.
            - selector (str, optional): The name of a predefined YAML selector that defines the set
                of resources to execute.
            - state (str, optional): The filepath of artifacts to use when establishing
                [state](https://docs.getdbt.com/docs/guides/understanding-state).
            - sync (bool, optional): Whether to run the command synchronously.
            - task_tags (dict, optional): Arbitrary key/value pairs to attach to this task. These
                tags will be returned in the output of the poll and ps methods.
            - threads (int, optional): The number of threads to use.
            - timeout (int, optional): The timeout in seconds.
            - timeout_action (str, optional): The action to take if the command times out. Only
                applies to sync mode.
        """

        exclude_str = " ".join(exclude) if exclude else None
        select_str = " ".join(select) if select else None
        task_tags = task_tags or {}

        params = {}
        if data is not None:
            params["data"] = data
        if exclude_str:
            params["exclude"] = exclude_str
        if schema is not None:
            params["schema"] = schema
        if select_str:
            params["select"] = select_str
        if selector:
            params["selector"] = selector
        if state:
            params["state"] = state
        if task_tags:
            params["task_tags"] = task_tags
        if threads:
            params["threads"] = threads
        if timeout:
            params["timeout"] = timeout

        if sync:
            return self._run_sync(
                "test",
                params=params,
                logs=logs,
                logs_start=logs_start,
                timeout=timeout,
                timeout_action=timeout_action,
                task_tags=task_tags,
            )
        return self._request("test", params=params)

    # pylint: disable=too-many-arguments
    def seed(
        self,
        *,
        exclude: List[str] = None,
        logs: bool = False,
        logs_start: int = 0,
        select: List[str] = None,
        selector: str = None,
        show: bool = None,
        state: str = None,
        sync: bool = False,
        task_tags: dict = None,
        threads: int = None,
        timeout: int = None,
        timeout_action: Union[str, Callable] = "raise",
    ):
        """
        Run seeds.

        Docs: https://docs.getdbt.com/reference/commands/rpc#run-seeds-docs

        Args:
            - exclude (list, optional): A list of resources to exclude from compiling, running,
                testing, seeding, or snapshotting .
            - logs (bool, optional): Whether to return logs. Only applies to sync mode.
            - logs_start (int, optional): The number of logs to skip. Only applies to sync mode.
            - select (list, optional): A list of resources to execute.
            - selector (str, optional): The name of a predefined YAML selector that defines the set
                of resources to execute.
            - show (bool, optional): If True, show the seed commands.
            - state (str, optional): The filepath of artifacts to use when establishing
                [state](https://docs.getdbt.com/docs/guides/understanding-state).
            - sync (bool, optional): Whether to run the command synchronously.
            - task_tags (dict, optional): Arbitrary key/value pairs to attach to this task. These
                tags will be returned in the output of the poll and ps methods.
            - threads (int, optional): The number of threads to use.
            - timeout (int, optional): The timeout in seconds.
            - timeout_action (str, optional): The action to take if the command times out. Only
                applies to sync mode.
        """

        exclude_str = " ".join(exclude) if exclude else None
        select_str = " ".join(select) if select else None
        task_tags = task_tags or {}

        params = {}
        if exclude_str:
            params["exclude"] = exclude_str
        if select_str:
            params["select"] = select_str
        if selector:
            params["selector"] = selector
        if show is not None:
            params["show"] = show
        if state:
            params["state"] = state
        if task_tags:
            params["task_tags"] = task_tags
        if threads:
            params["threads"] = threads
        if timeout:
            params["timeout"] = timeout

        if sync:
            return self._run_sync(
                "seed",
                params=params,
                logs=logs,
                logs_start=logs_start,
                timeout=timeout,
                timeout_action=timeout_action,
                task_tags=task_tags,
            )
        return self._request("seed", params=params)

    def snapshot(
        self,
        *,
        exclude: List[str] = None,
        logs: bool = False,
        logs_start: int = 0,
        select: List[str] = None,
        selector: str = None,
        state: str = None,
        sync: bool = False,
        task_tags: dict = None,
        threads: int = None,
        timeout: int = None,
        timeout_action: Union[str, Callable] = "raise",
    ):
        """
        Run snapshots.

        Docs: https://docs.getdbt.com/reference/commands/rpc#run-snapshots-docs

        Args:
            - exclude (list, optional): A list of resources to exclude from compiling, running,
                testing, seeding, or snapshotting .
            - logs (bool, optional): Whether to return logs. Only applies to sync mode.
            - logs_start (int, optional): The number of logs to skip. Only applies to sync mode.
            - select (list, optional): A list of resources to execute.
            - selector (str, optional): The name of a predefined YAML selector that defines the set
                of resources to execute.
            - state (str, optional): The filepath of artifacts to use when establishing
                [state](https://docs.getdbt.com/docs/guides/understanding-state).
            - sync (bool, optional): Whether to run the command synchronously.
            - task_tags (dict, optional): Arbitrary key/value pairs to attach to this task. These
                tags will be returned in the output of the poll and ps methods.
            - threads (int, optional): The number of threads to use.
            - timeout (int, optional): The timeout in seconds.
            - timeout_action (str, optional): The action to take if the command times out. Only
                applies to sync mode.
        """

        exclude_str = " ".join(exclude) if exclude else None
        select_str = " ".join(select) if select else None
        task_tags = task_tags or {}

        params = {}
        if exclude_str:
            params["exclude"] = exclude_str
        if select_str:
            params["select"] = select_str
        if selector:
            params["selector"] = selector
        if state:
            params["state"] = state
        if task_tags:
            params["task_tags"] = task_tags
        if threads:
            params["threads"] = threads
        if timeout:
            params["timeout"] = timeout

        if sync:
            return self._run_sync(
                "snapshot",
                params=params,
                logs=logs,
                logs_start=logs_start,
                timeout=timeout,
                timeout_action=timeout_action,
                task_tags=task_tags,
            )
        return self._request("snapshot", params=params)

    # pylint: disable=too-many-arguments
    def build(
        self,
        *,
        defer: str = None,
        exclude: List[str] = None,
        logs: bool = False,
        logs_start: int = 0,
        select: List[str] = None,
        selector: str = None,
        state: str = None,
        sync: bool = False,
        task_tags: dict = None,
        threads: int = None,
        timeout: int = None,
        timeout_action: Union[str, Callable] = "raise",
    ):
        """
        Build.

        Docs: https://docs.getdbt.com/reference/commands/rpc#build-docs

        Args:
            - defer (str, optional): ???
            - exclude (list, optional): A list of resources to exclude from compiling, running,
                testing, seeding, or snapshotting .
            - logs (bool, optional): Whether to return logs. Only applies to sync mode.
            - logs_start (int, optional): The number of logs to skip. Only applies to sync mode.
            - select (list, optional): A list of resources to execute.
            - selector (str, optional): The name of a predefined YAML selector that defines the set
                of resources to execute.
            - state (str, optional): The filepath of artifacts to use when establishing
                [state](https://docs.getdbt.com/docs/guides/understanding-state).
            - sync (bool, optional): Whether to run the command synchronously.
            - task_tags (dict, optional): Arbitrary key/value pairs to attach to this task. These
                tags will be returned in the output of the poll and ps methods.
            - threads (int, optional): The number of threads to use.
            - timeout (int, optional): The timeout in seconds.
            - timeout_action (str, optional): The action to take if the command times out. Only
                applies to sync mode.
        """

        exclude_str = " ".join(exclude) if exclude else None
        select_str = " ".join(select) if select else None
        task_tags = task_tags or {}

        params = {}
        if defer:
            params["defer"] = defer
        if exclude_str:
            params["exclude"] = exclude_str
        if select_str:
            params["select"] = select_str
        if selector:
            params["selector"] = selector
        if state:
            params["state"] = state
        if task_tags:
            params["task_tags"] = task_tags
        if threads:
            params["threads"] = threads
        if timeout:
            params["timeout"] = timeout

        if sync:
            return self._run_sync(
                "build",
                params=params,
                logs=logs,
                logs_start=logs_start,
                timeout=timeout,
                timeout_action=timeout_action,
                task_tags=task_tags,
            )
        return self._request("build", params=params)

    # pylint: disable=too-many-arguments
    def list_resources(
        self,
        *,
        exclude: List[str] = None,
        logs: bool = False,
        logs_start: int = 0,
        output_keys: List[str] = None,
        resource_types: List[str] = None,
        select: List[str] = None,
        selector: str = None,
        sync: bool = False,
        task_tags: dict = None,
        threads: int = None,
        timeout: int = None,
        timeout_action: Union[str, Callable] = "raise",
    ):
        """
        List project resources.

        Docs: https://docs.getdbt.com/reference/commands/rpc#list-project-resources-docs

        Args:
            - exclude (list, optional): A list of resources to exclude from compiling, running,
                testing, seeding, or snapshotting .
            - logs (bool, optional): Whether to return logs. Only applies to sync mode.
            - logs_start (int, optional): The number of logs to skip. Only applies to sync mode.
            - output_keys (list, optional): Specify which node properties to include in output.
            - resource_types (list, optional): Filter selected resources by type.
            - select (list, optional): A list of resources to execute.
            - selector (str, optional): The name of a predefined YAML selector that defines the set
                of resources to execute.
            - sync (bool, optional): Whether to run the command synchronously.
            - task_tags (dict, optional): Arbitrary key/value pairs to attach to this task. These
                tags will be returned in the output of the poll and ps methods.
            - threads (int, optional): The number of threads to use.
            - timeout (int, optional): The timeout in seconds.
            - timeout_action (str, optional): The action to take if the command times out. Only
                applies to sync mode.
        """

        exclude_str = " ".join(exclude) if exclude else None
        select_str = " ".join(select) if select else None
        task_tags = task_tags or {}

        params = {}
        if exclude_str:
            params["exclude"] = exclude_str
        if output_keys:
            params["output_keys"] = output_keys
        if resource_types:
            params["resource_types"] = resource_types
        if select_str:
            params["select"] = select_str
        if selector:
            params["selector"] = selector
        if task_tags:
            params["task_tags"] = task_tags
        if threads:
            params["threads"] = threads
        if timeout:
            params["timeout"] = timeout

        if sync:
            return self._run_sync(
                "build",
                params=params,
                logs=logs,
                logs_start=logs_start,
                timeout=timeout,
                timeout_action=timeout_action,
                task_tags=task_tags,
            )
        return self._request("build", params=params)

    # pylint: disable=too-many-arguments
    def generate_docs(
        self,
        *,
        compile_project: bool = None,
        logs: bool = False,
        logs_start: int = 0,
        state: str = None,
        sync: bool = False,
        task_tags: dict = None,
        timeout: int = None,
        timeout_action: Union[str, Callable] = "raise",
    ):
        """
        Generate docs.

        Docs: https://docs.getdbt.com/reference/commands/rpc#generate-docs-docs

        Args:
            - compile_project (bool, optional): If True, compile the project before generating a
                catalog (optional, default=false).
            - logs (bool, optional): Whether to return logs. Only applies to sync mode.
            - logs_start (int, optional): The number of logs to skip. Only applies to sync mode.
            - state (str, optional): The filepath of artifacts to use when establishing
                [state](https://docs.getdbt.com/docs/guides/understanding-state).
            - sync (bool, optional): Whether to run the command synchronously.
            - task_tags (dict, optional): Arbitrary key/value pairs to attach to this task. These
                tags will be returned in the output of the poll and ps methods.
            - timeout (int, optional): The timeout in seconds.
            - timeout_action (str, optional): The action to take if the command times out. Only
                applies to sync mode.
        """

        task_tags = task_tags or {}

        params = {}
        if compile_project:
            params["compile"] = compile_project
        if state:
            params["state"] = state
        if task_tags:
            params["task_tags"] = task_tags

        if sync:
            return self._run_sync(
                "docs.generate",
                params=params,
                logs=logs,
                logs_start=logs_start,
                timeout=timeout,
                timeout_action=timeout_action,
                task_tags=task_tags,
            )
        return self._request("docs.generate", params=params)

    def compile_sql(
        self,
        name: str,
        sql: str,
        *,
        logs: bool = False,
        logs_start: int = 0,
        sync: bool = False,
        task_tags: dict = None,
        timeout: int = 60,
        timeout_action: Union[str, Callable] = "raise",
    ):
        """
        Compiles a SQL statement.

        Docs: https://docs.getdbt.com/reference/commands/rpc#compiling-a-query

        Args:
            - logs (bool, optional): Whether to return logs. Only applies to sync mode.
            - logs_start (int, optional): The number of logs to skip. Only applies to sync mode.
            - name (str): The name of the query.
            - sql (str): The SQL statement to compile. No need to encode it in base64 before.
            - sync (bool, optional): Whether to run the command synchronously.
            - task_tags (dict, optional): Arbitrary key/value pairs to attach to this task. These
                tags will be returned in the output of the poll and ps methods.
            - timeout (int, optional): The timeout in seconds.
            - timeout_action (str, optional): The action to take if the command times out.
                Only applies to sync mode.
        """

        task_tags = task_tags or {}

        params = {}
        if task_tags:
            params["task_tags"] = task_tags
        if timeout:
            params["timeout"] = timeout
        params["name"] = name
        params["sql"] = base64.b64encode(sql.encode("utf-8")).decode("utf-8")

        if sync:
            return self._run_sync(
                "compile_sql",
                params=params,
                logs=logs,
                logs_start=logs_start,
                timeout=timeout,
                timeout_action=timeout_action,
                task_tags=task_tags,
            )
        return self._request("compile_sql", params=params)

    def run_sql(
        self,
        name: str,
        sql: str,
        *,
        logs: bool = False,
        logs_start: int = 0,
        sync: bool = False,
        task_tags: dict = None,
        timeout: int = 60,
        timeout_action: Union[str, Callable] = "raise",
    ):
        """
        Runs a SQL statement.

        Docs: https://docs.getdbt.com/reference/commands/rpc#executing-a-query

        Args:
            - logs (bool, optional): Whether to return logs. Only applies to sync mode.
            - logs_start (int, optional): The number of logs to skip. Only applies to sync mode.
            - name (str): The name of the query.
            - sql (str): The SQL statement to run. No need to encode it in base64 before.
            - sync (bool, optional): Whether to run the command synchronously.
            - task_tags (dict, optional): Arbitrary key/value pairs to attach to this task. These
                tags will be returned in the output of the poll and ps methods.
            - timeout (int, optional): The timeout in seconds.
            - timeout_action (str, optional): The action to take if the command times out. Only
                applies to sync mode.
        """

        task_tags = task_tags or {}

        params = {}
        if task_tags:
            params["task_tags"] = task_tags
        if timeout:
            params["timeout"] = timeout
        params["name"] = name
        params["sql"] = base64.b64encode(sql.encode("utf-8")).decode("utf-8")

        if sync:
            return self._run_sync(
                "run_sql",
                params=params,
                logs=logs,
                logs_start=logs_start,
                timeout=timeout,
                timeout_action=timeout_action,
                task_tags=task_tags,
            )
        return self._request("run_sql", params=params)



"""
A super lightweight abstraction of the dbt rpc which serves out synchronous requests
and affords more customization. Should also afford us the ability to swap the RPC
with another solution like `dbt.lib` which dbt server may implicitly use.
"""
import multiprocessing
import subprocess
import time
from typing import Dict

import click
from flask import Flask, request

from .dbt_rpc_client import DbtClient, RPCError

app = Flask(__name__)
__version__ = "0.2.6"

STATE: Dict[str, DbtClient] = {}

LOG_MSG = """
{action} QUERY
===============
{query}
"""


@app.route("/run", methods=["POST"])
def run_sql():
    # Server Logging
    print(LOG_MSG.format(action="RUNNING", query=request.data))
    try:
        # TODO: Lets consider memoization, also make the limit parameterizable
        result = STATE["server"].run_sql(
            "dbt-sync-server",
            f'SELECT * FROM ({request.data.decode("UTF-8")}) AS __rpc_query LIMIT 200',
            sync=True,
        )
    except RPCError as rpc_err:
        return rpc_err.response
    else:
        return {
            **result["result"]["results"][0]["table"],
            "compiled_sql": result["result"]["results"][0]["compiled_sql"],
            "raw_sql": result["result"]["results"][0]["raw_sql"],
        }


@app.route("/compile", methods=["POST"])
def compile_sql():
    # Server Logging
    print(LOG_MSG.format(action="COMPILING", query=request.data))
    try:
        # Lets consider memoization
        result = STATE["server"].compile_sql(
            "dbt-sync-server", request.data.decode("UTF-8"), sync=True
        )
    except RPCError as rpc_err:
        return rpc_err.response
    else:
        return {"result": result["result"]["results"][0]["compiled_sql"]}


@app.route("/api/health", methods=["GET"])
def health_check(raise_on_error: bool = False) -> Dict[str, str]:
    """Example response
    {
    "result": {
        "status": "ready",
        "error": null,
        "logs": [..],
        "timestamp": "2019-10-07T16:30:09.875534Z",
        "pid": 76715
    },
    "id": "2db9a2fe-9a39-41ef-828c-25e04dd6b07d",
    "jsonrpc": "2.0"
    }"""
    try:
        result = STATE["server"].status()
    except RPCError as rpc_err:
        if raise_on_error:
            raise ConnectionError from rpc_err
        return rpc_err.response
    except Exception as exc:
        # Catch alternate errors
        if raise_on_error:
            raise ConnectionError from exc
        return {"error": f"Unknown error has occured: {str(exc)}"}
    else:
        return result


def run_rpc(
    rpc_port: int, project_dir: str, profiles_dir: str, profile: str, target: str
):
    print(
        f"Starting RPC port:{rpc_port} project_dir:{project_dir} profiles_dir:{profiles_dir} profile:{profile} target:{target}"
    )  # noqa
    try:
        with open("dbt_rpc.log", "w") as f:
            subprocess.run(
                [
                    "dbt-rpc",
                    "serve",
                    "--port",
                    str(rpc_port),
                    "--project-dir",
                    str(project_dir),
                    "--profiles-dir",
                    str(profiles_dir),
                    "--profile",
                    str(profile),
                    "--target",
                    str(target),
                ],
                stdout=f,
                stderr=subprocess.STDOUT,
            )
    except Exception as err:
        print("RPC Terminated? Error: {}".format(str(err)))


@click.group()
@click.version_option(__version__)
def cli():
    pass


@cli.command()
@click.option("--port", type=click.INT, default=8581)
@click.option("--rpc-port", type=click.INT, default=8580)
@click.option(
    "--project-dir",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    default="./",
)
@click.option(
    "--profiles-dir",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    default="~/.dbt",
)
@click.option("--profile", type=click.STRING, default="default")
@click.option("--target", type=click.STRING, default="dev")
@click.option("--no-inject-rpc", is_flag=True, type=click.BOOL, default=False)
def serve(
    port: int = 8581,
    rpc_port: int = 8580,
    project_dir: str = "./",
    profiles_dir: str = "~/.dbt",
    profile: str = "default",
    target: str = "dev",
    no_inject_rpc: bool = False,
):
    STATE["server"] = DbtClient(port=rpc_port)
    if not no_inject_rpc:
        rpc_server = multiprocessing.Process(
            target=run_rpc,
            args=(rpc_port, project_dir, profiles_dir, profile, target),
            daemon=True,
        )
        rpc_server.start()
        time.sleep(2.5)
        if not rpc_server.is_alive():
            exit_code = rpc_server.exitcode
            rpc_server.close()
            if exit_code == 0:
                print(
                    "RPC failed to initialize, exit code {} most likely indicates a process is already running on port {} or the project directory provided [{}] is not a valid dbt project.".format(
                        exit_code, rpc_port, project_dir
                    )
                )
            elif exit_code == 1:
                print(
                    "RPC failed to initialize, exit code {} most likely indicates the dbt project is invalid or has an error.".format(
                        exit_code
                    )
                )
            else:
                print(
                    "RPC failed to initialize, exit code {} with unknown root cause.".format(
                        exit_code
                    )
                )
            exit(1)
    try:
        # ping RPC to gaurantee connectivity before starting flask app
        ping_count = 0
        while ping_count < 3:
            if not health_check().get("error"):
                print("RPC health check passed!")
                break
            ping_count += 1
            time.sleep(2.5)
        else:
            print("RPC health check failing, final attempt")
            health_check(raise_on_error=True)
        app.run("localhost", port)
    finally:
        print("\nSHUTDOWN")
        if not no_inject_rpc and rpc_server.is_alive():
            print("CLEANING UP RPC")
            rpc_server.terminate()
            rpc_server.join()
            rpc_server.close()


@cli.command()
def build_shell_scripts():
    """Build shell scripts in local directory. These handle spinning up the server and
    setting up a cron job which will reparse dbt project as needed. Invoke the from the
    main dbt project directory. It includes a kill server script too which will clean up both the
    RPC and the wrapper."""
    # run_server = Path(__file__).parent.parent.parent / "bin" / "run_server.sh"
    # kill_server = Path(__file__).parent.parent.parent / "bin" / "kill_server.sh"
    from textwrap import dedent

    with open("./kill_server.sh", "w") as run_server_target:
        run_server_target.write(
            dedent(
                """
            #!/bin/bash

            # RUN SERVER
            dbt-sync-server serve > dbt_sync.log 2>&1 &

            # CAPTURE PIDs
            echo $! >/tmp/dbt_sync_server.pid
            sleep 2.5
            ps aux | grep dbt-rpc | grep -v grep | awk '{print $2}' >/tmp/dbt_rpc_server.pid

            # SET CRON VARS
            dbt_rpc_id=`cat /tmp/dbt_rpc_server.pid`
            every_x_minutes=5

            # REPARSE DBT PROJECT EVERY X MINUTES
            crontab -l | grep -v '__dbt reparsed__' | crontab
            (crontab -l ; echo "1/$every_x_minutes * * * * kill -HUP $dbt_rpc_id ; echo '__dbt reparsed__'") | crontab

            # USE THIS /tmp/dbt_rpc_server.pid TO REPARSE MANUALLY
            echo $dbt_rpc_id

        """
            )
        )
    with open("./run_server.sh", "w") as kill_server_target:
        kill_server_target.write(
            dedent(
                """
            #!/bin/bash

            # SET VARS
            dbt_rpc_id=`cat /tmp/dbt_rpc_server.pid`
            dbt_ss_id=`cat /tmp/dbt_sync_server.pid`

            # REPARSE DBT PROJECT EVERY X MINUTES
            crontab -l | grep -v '__dbt reparsed__' | crontab
            kill $dbt_rpc_id ; kill $dbt_ss_id

        """
            )
        )


if __name__ == "__main__":
    cli()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['dbt_sync_server']

package_data = \
{'': ['*']}

install_requires = \
['Flask>=1,<2',
 'Werkzeug<2',
 'click>=8.0.4,<9.0.0',
 'dbt-core>1.0.1',
 'dbt-rpc>=0.1.1,<0.2.0',
 'itsdangerous==2.0.1']

entry_points = \
{'console_scripts': ['dbt-sync-server = dbt_sync_server:cli']}

setup_kwargs = {
    'name': 'dbt-sync-server',
    'version': '0.2.8',
    'description': 'Server which abstracts RPC and simplifies SQL execution and compilation',
    'long_description': None,
    'author': 'Alex Butler',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)
